@@ -1,7 +1,7 @@
 package Bencher;
 
-our $DATE = '2016-01-11'; # DATE
-our $VERSION = '0.29'; # VERSION
+our $DATE = '2016-01-19'; # DATE
+our $VERSION = '0.30'; # VERSION
 
 use 5.010001;
 use strict;
@@ -883,95 +883,66 @@ $SPEC{format_result} = {
             req => 1,
             pos => 0,
         },
+        formatters => {
+            summary => 'Formatters specification',
+            schema => ['array*', of=>[
+                'any*', of=>[
+                    'str*',
+                    ['array*', len=>2, elems=>['str*', 'hash*']],
+                ]
+            ]],
+            req => 1,
+            pos => 1,
+        },
     },
     args_as => 'array',
 };
 sub format_result {
     require POSIX;
 
-    my $envres = shift;
+    my ($envres, $formatters) = @_;
 
-    my $ff = $envres->[3]{'table.fields'};
+    $formatters //= [
+        'Sort',
+        'ScaleTime',
+        'ScaleRate',
+        'RoundNumbers',
+        ($envres->[3]{'func.module_startup'} ? ('ModuleStartup') : ()),
+        'DeleteConstantFields',
+        'DeleteNotesFieldIfEmpty',
+        'DeleteSeqField',
 
-    # for module_startup mode: remove 'rate', add 'mod_overhead_time'
-    if ($envres->[3]{'func.module_startup'}) {
-        my $rit_baseline = _find_record_by_seq($envres->[2], 0);
-        for my $rit (@{$envres->[2]}) {
-            delete $rit->{rate};
-            if ($rit_baseline) {
-                $rit->{mod_overhead_time} =
-                    $rit->{time} - $rit_baseline->{time};
-            }
+        'RenderAsTextTable',
+    ];
+
+    # load all formatter modules
+    my @fmtobjs;
+    for my $fmt (@$formatters) {
+        my ($fmtname, $fmtargs);
+        if (ref($fmt)) {
+            $fmtname = $fmt->[0];
+            $fmtargs = $fmt->[1];
+        } else {
+            $fmtname = $fmt;
+            $fmtargs = {};
         }
-        splice @$ff, (firstidx {$_ eq 'rate'} @$ff), 1;
-        splice @$ff, (firstidx {$_ eq 'time'} @$ff)+1, 0, "mod_overhead_time";
+        my $fmtmod = "Bencher::Formatter::$fmtname";
+        my $fmtmod_pm = $fmtmod; $fmtmod_pm =~ s!::!/!g; $fmtmod_pm .= ".pm";
+        require $fmtmod_pm;
+        push @fmtobjs, $fmtmod->new(%$fmtargs);
     }
 
-    # sort by default from slowest to fastest
-    $envres->[2] = [sort {$b->{time} <=> $a->{time}} @{$envres->[2]}];
-
-    # pick an appropriate time unit & format the time
-    my ($min, $max) = minmax(map {$_->{time}} @{$envres->[2]});
-
-    my ($unit, $factor);
-    if ($max <= 1.5e-6) {
-        $unit = "ns";
-        $factor = 1e9;
-    } elsif ($max <= 1.5e-3) {
-        $unit = "\x{03bc}s"; # XXX enable utf
-        $factor = 1e6;
-    } elsif ($max <= 1.5) {
-        $unit = "ms";
-        $factor = 1e3;
+    # run all munge_result()
+    for my $fmtobj (@fmtobjs) {
+        next unless $fmtobj->can("munge_result");
+        $fmtobj->munge_result($envres);
     }
 
-    if ($unit) {
-        for my $rit (@{$envres->[2]}) {
-            my $num_significant_digits =
-                $rit->{errors} == 0 ? 6 :
-                POSIX::round(log($rit->{time} / $rit->{errors})/log(10));
-            $rit->{time} = sprintf(
-                "%.${num_significant_digits}g%s",
-                $rit->{time} * $factor, $unit);
-            if (exists $rit->{rate}) {
-                $rit->{rate} = sprintf(
-                    "%.${num_significant_digits}g", $rit->{rate});
-            }
-            $rit->{errors} = sprintf("%.2g", $rit->{errors});
-            if (exists $rit->{mod_overhead_time}) {
-                $rit->{mod_overhead_time} = sprintf(
-                    "%.${num_significant_digits}g%s",
-                    $rit->{mod_overhead_time} * $factor, $unit);
-            }
-        }
+    # return the first render_result()
+    for my $fmtobj (@fmtobjs) {
+        next unless $fmtobj->can("render_result");
+        return $fmtobj->render_result($envres);
     }
-
-    # remove constant item permutation columns to reduce clutter
-    {
-        last unless @{$envres->[2]};
-        require TableData::Object::aohos;
-        my $td = TableData::Object::aohos->new($envres->[2]);
-        my @const_cols = $td->const_col_names;
-        for my $k (@const_cols) {
-            next unless $k =~ /^(item_.+|arg_.+|participant|dataset)$/;
-            for my $row (@{ $envres->[2] }) {
-                delete $row->{$k};
-            }
-        }
-    }
-
-    # remove notes column if there are none
-  REMOVE_NOTES:
-    {
-        for my $row (@{$envres->[2]}) {
-            last REMOVE_NOTES if $row->{notes};
-        }
-        for my $row (@{$envres->[2]}) {
-            delete $row->{notes};
-        }
-    }
-
-    $envres;
 }
 
 $SPEC{bencher} = {
@@ -1765,7 +1736,13 @@ sub bencher {
             my $r = $args{-cmdline_r};
             last unless $r && ($r->{format} // 'text') =~ /text/;
 
-            format_result($envres);
+            $envres = [
+                200, "OK",
+                format_result($envres),
+                {
+                    "cmdline.skip_format" => 1,
+                },
+            ];
         }
 
         if ($return_resmeta) {
@@ -1829,7 +1806,7 @@ Bencher - A benchmark framework
 
 =head1 VERSION
 
-This document describes version 0.29 of Bencher (from Perl distribution Bencher), released on 2016-01-11.
+This document describes version 0.30 of Bencher (from Perl distribution Bencher), released on 2016-01-19.
 
 =head1 SYNOPSIS
 
@@ -1861,271 +1838,6 @@ on.
 
 =back
 
-=head1 FUNCTIONS
-
-
-=head2 bencher(%args) -> [status, msg, result, meta]
-
-A benchmark framework.
-
-This function is not exportable.
-
-Arguments ('*' denotes required arguments):
-
-=over 4
-
-=item * B<action> => I<str> (default: "bench")
-
-=item * B<datasets> => I<array[hash]>
-
-Add datasets.
-
-=item * B<detail> => I<bool>
-
-=item * B<exclude_dataset_pattern> => I<re>
-
-Exclude datasets matching this regex pattern.
-
-=item * B<exclude_dataset_tags> => I<array[str]>
-
-Exclude datasets whose tag matches this.
-
-You can specify C<A & B> to exclude datasets that have I<both> tags A and B.
-
-=item * B<exclude_datasets> => I<array[str]>
-
-Exclude datasets whose seq/name matches this.
-
-=item * B<exclude_function_pattern> => I<re>
-
-Exclude function(s) matching this regex pattern.
-
-=item * B<exclude_functions> => I<array[str]>
-
-Exclude functions specified in this list.
-
-=item * B<exclude_item_pattern> => I<re>
-
-Exclude items matching this regex pattern.
-
-=item * B<exclude_items> => I<array[str]>
-
-Exclude items whose seq/name matches this.
-
-=item * B<exclude_module_pattern> => I<re>
-
-Exclude module(s) matching this regex pattern.
-
-=item * B<exclude_modules> => I<array[str]>
-
-Exclude modules specified in this list.
-
-=item * B<exclude_participant_pattern> => I<re>
-
-Exclude participants matching this regex pattern.
-
-=item * B<exclude_participant_tags> => I<array[str]>
-
-Exclude participants whose tag matches this.
-
-You can specify C<A & B> to exclude participants that have I<both> tags A and B.
-
-=item * B<exclude_participants> => I<array[str]>
-
-Exclude participants whose seq/name matches this.
-
-=item * B<include_dataset_pattern> => I<re>
-
-Only include datasets matching this regex pattern.
-
-=item * B<include_dataset_tags> => I<array[str]>
-
-Only include datasets whose tag matches this.
-
-You can specify C<A & B> to include datasets that have I<both> tags A and B.
-
-=item * B<include_datasets> => I<array[str]>
-
-Only include datasets whose seq/name matches this.
-
-=item * B<include_function_pattern> => I<re>
-
-Only include functions matching this regex pattern.
-
-=item * B<include_functions> => I<array[str]>
-
-Only include functions specified in this list.
-
-=item * B<include_item_pattern> => I<re>
-
-Only include items matching this regex pattern.
-
-=item * B<include_items> => I<array[str]>
-
-Only include items whose seq/name matches this.
-
-=item * B<include_module_pattern> => I<re>
-
-Only include modules matching this regex pattern.
-
-=item * B<include_modules> => I<array[str]>
-
-Only include modules specified in this list.
-
-=item * B<include_participant_pattern> => I<re>
-
-Only include participants matching this regex pattern.
-
-=item * B<include_participant_tags> => I<array[str]>
-
-Only include participants whose tag matches this.
-
-You can specify C<A & B> to include participants that have I<both> tags A and B.
-
-=item * B<include_participants> => I<array[str]>
-
-Only include participants whose seq/name matches this.
-
-=item * B<module_startup> => I<bool>
-
-Benchmark module startup overhead instead of normal benchmark.
-
-=item * B<on_failure> => I<str>
-
-What to do when there is a failure.
-
-For a command participant, failure means non-zero exit code. For a Perl-code
-participant, failure means Perl code dies or (if expected result is specified)
-the result is not equal to the expected result.
-
-The default is "die". When set to "skip", will first run the code of each item
-before benchmarking and trap command failure/Perl exception and if that happens,
-will "skip" the item.
-
-=item * B<on_result_failure> => I<str>
-
-What to do when there is a result failure.
-
-This is like C<on_failure> except that it specifically refer to the failure of
-item's result not being equal to expected result.
-
-There is an extra choice of C<warn> for this type of failure, which is to print a
-warning to STDERR and continue.
-
-=item * B<participants> => I<array[hash]>
-
-Add participants.
-
-=item * B<precision> => I<float>
-
-Precision, will be passed to Benchmark::Dumb.
-
-This setting overrides C<default_precision> property in the scenario.
-
-=item * B<precision_limit> => I<float>
-
-Set maximum (=smallest number) precision.
-
-Instead of setting C<precision> which forces a single value, you can also set
-this C<precision_limit> setting. If the precision in the scenario is higher
-(=number is smaller) than this limit, then this limit is used. For example, if
-the scenario specifies C<default_precision> 0.001 and C<precision_limit> is set to
-0.005 then 0.005 is used.
-
-This setting is useful on slower computers which might not be able to reach the
-required precision before hitting maximum number of iterations.
-
-=item * B<raw> => I<bool>
-
-Show "raw" data.
-
-When action=show-items-result, will print result as-is instead of dumping as
-Perl.
-
-=item * B<scenario_file> => I<str>
-
-Load a scenario from a Perl file.
-
-Perl file will be do()'ed and the last expression should be a hash containing
-the scenario specification.
-
-=item * B<scenario_module> => I<str>
-
-Load a scenario from a Bencher::Scenario:: Perl module.
-
-Will try to load module C<< Bencher::Scenario::E<lt>NAMEE<gt> >> and expect to find a package
-variable in the module called C<$scenario> which should be a hashref containing
-the scenario specification.
-
-=back
-
-Returns an enveloped result (an array).
-
-First element (status) is an integer containing HTTP status code
-(200 means OK, 4xx caller error, 5xx function error). Second element
-(msg) is a string containing error message, or 'OK' if status is
-200. Third element (result) is optional, the actual result. Fourth
-element (meta) is called result metadata and is optional, a hash
-that contains extra information.
-
-Return value:  (any)
-
-
-=head2 format_result($envres) -> [status, msg, result, meta]
-
-Format bencher result.
-
-This function is not exportable.
-
-Arguments ('*' denotes required arguments):
-
-=over 4
-
-=item * B<envres>* => I<array>
-
-Enveloped result from bencher.
-
-=back
-
-Returns an enveloped result (an array).
-
-First element (status) is an integer containing HTTP status code
-(200 means OK, 4xx caller error, 5xx function error). Second element
-(msg) is a string containing error message, or 'OK' if status is
-200. Third element (result) is optional, the actual result. Fourth
-element (meta) is called result metadata and is optional, a hash
-that contains extra information.
-
-Return value:  (any)
-
-
-=head2 parse_scenario(%args) -> [status, msg, result, meta]
-
-Parse scenario (fill in default values, etc).
-
-This function is not exportable.
-
-Arguments ('*' denotes required arguments):
-
-=over 4
-
-=item * B<scenario> => I<hash>
-
-Unparsed scenario.
-
-=back
-
-Returns an enveloped result (an array).
-
-First element (status) is an integer containing HTTP status code
-(200 means OK, 4xx caller error, 5xx function error). Second element
-(msg) is a string containing error message, or 'OK' if status is
-200. Third element (result) is optional, the actual result. Fourth
-element (meta) is called result metadata and is optional, a hash
-that contains extra information.
-
-Return value:  (any)
-
 =head1 SCENARIO
 
 The core data structure that you need to prepare is the B<scenario>. It is a
@@ -2380,18 +2092,274 @@ You can use this hook to, e.g.: modify the result in some way.
 
 =back
 
-=head1 SEE ALSO
+=head1 FUNCTIONS
 
-L<bencher>
 
-B<BenchmarkAnything>. There are lot of overlaps of goals between Bencher and
-this project. I hope to reuse or interoperate parts of BenchmarkAnything, e.g.
-storing Bencher results in a BenchmarkAnything storage backend, sending Bencher
-results to a BenchmarkAnything HTTP server, and so on.
+=head2 bencher(%args) -> [status, msg, result, meta]
 
-L<Benchmark>, L<Benchmark::Dumb> (L<Dumbbench>)
+A benchmark framework.
 
-C<Bencher::Scenario::*> for examples of scenarios.
+This function is not exportable.
+
+Arguments ('*' denotes required arguments):
+
+=over 4
+
+=item * B<action> => I<str> (default: "bench")
+
+=item * B<datasets> => I<array[hash]>
+
+Add datasets.
+
+=item * B<detail> => I<bool>
+
+=item * B<exclude_dataset_pattern> => I<re>
+
+Exclude datasets matching this regex pattern.
+
+=item * B<exclude_dataset_tags> => I<array[str]>
+
+Exclude datasets whose tag matches this.
+
+You can specify C<A & B> to exclude datasets that have I<both> tags A and B.
+
+=item * B<exclude_datasets> => I<array[str]>
+
+Exclude datasets whose seq/name matches this.
+
+=item * B<exclude_function_pattern> => I<re>
+
+Exclude function(s) matching this regex pattern.
+
+=item * B<exclude_functions> => I<array[str]>
+
+Exclude functions specified in this list.
+
+=item * B<exclude_item_pattern> => I<re>
+
+Exclude items matching this regex pattern.
+
+=item * B<exclude_items> => I<array[str]>
+
+Exclude items whose seq/name matches this.
+
+=item * B<exclude_module_pattern> => I<re>
+
+Exclude module(s) matching this regex pattern.
+
+=item * B<exclude_modules> => I<array[str]>
+
+Exclude modules specified in this list.
+
+=item * B<exclude_participant_pattern> => I<re>
+
+Exclude participants matching this regex pattern.
+
+=item * B<exclude_participant_tags> => I<array[str]>
+
+Exclude participants whose tag matches this.
+
+You can specify C<A & B> to exclude participants that have I<both> tags A and B.
+
+=item * B<exclude_participants> => I<array[str]>
+
+Exclude participants whose seq/name matches this.
+
+=item * B<include_dataset_pattern> => I<re>
+
+Only include datasets matching this regex pattern.
+
+=item * B<include_dataset_tags> => I<array[str]>
+
+Only include datasets whose tag matches this.
+
+You can specify C<A & B> to include datasets that have I<both> tags A and B.
+
+=item * B<include_datasets> => I<array[str]>
+
+Only include datasets whose seq/name matches this.
+
+=item * B<include_function_pattern> => I<re>
+
+Only include functions matching this regex pattern.
+
+=item * B<include_functions> => I<array[str]>
+
+Only include functions specified in this list.
+
+=item * B<include_item_pattern> => I<re>
+
+Only include items matching this regex pattern.
+
+=item * B<include_items> => I<array[str]>
+
+Only include items whose seq/name matches this.
+
+=item * B<include_module_pattern> => I<re>
+
+Only include modules matching this regex pattern.
+
+=item * B<include_modules> => I<array[str]>
+
+Only include modules specified in this list.
+
+=item * B<include_participant_pattern> => I<re>
+
+Only include participants matching this regex pattern.
+
+=item * B<include_participant_tags> => I<array[str]>
+
+Only include participants whose tag matches this.
+
+You can specify C<A & B> to include participants that have I<both> tags A and B.
+
+=item * B<include_participants> => I<array[str]>
+
+Only include participants whose seq/name matches this.
+
+=item * B<module_startup> => I<bool>
+
+Benchmark module startup overhead instead of normal benchmark.
+
+=item * B<on_failure> => I<str>
+
+What to do when there is a failure.
+
+For a command participant, failure means non-zero exit code. For a Perl-code
+participant, failure means Perl code dies or (if expected result is specified)
+the result is not equal to the expected result.
+
+The default is "die". When set to "skip", will first run the code of each item
+before benchmarking and trap command failure/Perl exception and if that happens,
+will "skip" the item.
+
+=item * B<on_result_failure> => I<str>
+
+What to do when there is a result failure.
+
+This is like C<on_failure> except that it specifically refer to the failure of
+item's result not being equal to expected result.
+
+There is an extra choice of C<warn> for this type of failure, which is to print a
+warning to STDERR and continue.
+
+=item * B<participants> => I<array[hash]>
+
+Add participants.
+
+=item * B<precision> => I<float>
+
+Precision, will be passed to Benchmark::Dumb.
+
+This setting overrides C<default_precision> property in the scenario.
+
+=item * B<precision_limit> => I<float>
+
+Set maximum (=smallest number) precision.
+
+Instead of setting C<precision> which forces a single value, you can also set
+this C<precision_limit> setting. If the precision in the scenario is higher
+(=number is smaller) than this limit, then this limit is used. For example, if
+the scenario specifies C<default_precision> 0.001 and C<precision_limit> is set to
+0.005 then 0.005 is used.
+
+This setting is useful on slower computers which might not be able to reach the
+required precision before hitting maximum number of iterations.
+
+=item * B<raw> => I<bool>
+
+Show "raw" data.
+
+When action=show-items-result, will print result as-is instead of dumping as
+Perl.
+
+=item * B<scenario_file> => I<str>
+
+Load a scenario from a Perl file.
+
+Perl file will be do()'ed and the last expression should be a hash containing
+the scenario specification.
+
+=item * B<scenario_module> => I<str>
+
+Load a scenario from a Bencher::Scenario:: Perl module.
+
+Will try to load module C<< Bencher::Scenario::E<lt>NAMEE<gt> >> and expect to find a package
+variable in the module called C<$scenario> which should be a hashref containing
+the scenario specification.
+
+=back
+
+Returns an enveloped result (an array).
+
+First element (status) is an integer containing HTTP status code
+(200 means OK, 4xx caller error, 5xx function error). Second element
+(msg) is a string containing error message, or 'OK' if status is
+200. Third element (result) is optional, the actual result. Fourth
+element (meta) is called result metadata and is optional, a hash
+that contains extra information.
+
+Return value:  (any)
+
+
+=head2 format_result($envres, $formatters) -> [status, msg, result, meta]
+
+Format bencher result.
+
+This function is not exportable.
+
+Arguments ('*' denotes required arguments):
+
+=over 4
+
+=item * B<envres>* => I<array>
+
+Enveloped result from bencher.
+
+=item * B<formatters>* => I<array[str|array]>
+
+Formatters specification.
+
+=back
+
+Returns an enveloped result (an array).
+
+First element (status) is an integer containing HTTP status code
+(200 means OK, 4xx caller error, 5xx function error). Second element
+(msg) is a string containing error message, or 'OK' if status is
+200. Third element (result) is optional, the actual result. Fourth
+element (meta) is called result metadata and is optional, a hash
+that contains extra information.
+
+Return value:  (any)
+
+
+=head2 parse_scenario(%args) -> [status, msg, result, meta]
+
+Parse scenario (fill in default values, etc).
+
+This function is not exportable.
+
+Arguments ('*' denotes required arguments):
+
+=over 4
+
+=item * B<scenario> => I<hash>
+
+Unparsed scenario.
+
+=back
+
+Returns an enveloped result (an array).
+
+First element (status) is an integer containing HTTP status code
+(200 means OK, 4xx caller error, 5xx function error). Second element
+(msg) is a string containing error message, or 'OK' if status is
+200. Third element (result) is optional, the actual result. Fourth
+element (meta) is called result metadata and is optional, a hash
+that contains extra information.
+
+Return value:  (any)
 
 =head1 HOMEPAGE
 
@@ -2409,6 +2377,19 @@ When submitting a bug or request, please include a test-file or a
 patch to an existing test-file that illustrates the bug or desired
 feature.
 
+=head1 SEE ALSO
+
+L<bencher>
+
+B<BenchmarkAnything>. There are lot of overlaps of goals between Bencher and
+this project. I hope to reuse or interoperate parts of BenchmarkAnything, e.g.
+storing Bencher results in a BenchmarkAnything storage backend, sending Bencher
+results to a BenchmarkAnything HTTP server, and so on.
+
+L<Benchmark>, L<Benchmark::Dumb> (L<Dumbbench>)
+
+C<Bencher::Scenario::*> for examples of scenarios.
+
 =head1 AUTHOR
 
 perlancar <perlancar@cpan.org>
